#include "rtm_zero_op_3d.h"
#include "sinc_bob.h"
#include "gpu_funcs_3d.h"
#include <math.h>
#include "deriv_3d.h"
#include "map_data_3d.h"
rtm_zero_op::rtm_zero_op(param_func *par,vel_fd_3d *vel_3d, source_func *source_func, data_rtm_3d *dat,data_rtm_3d *datin, image_rtm_3d *img,float ap,bool v,bool enc,int *r_vec, bool do_src,bool redo_src){
   std::vector<float> ds; 
   data=dat;
   bdata=datin;
   ds.push_back(dat->dt);
   ds.push_back(source_func->get_dt());
   set_vel(vel_3d);
   calc_stability(&ds,data->nt);
   set_fd_basics(par,source_func,ap,v);
   image=img;
   set_domain(image);
   set_range(data);
   dtd=data->dt;
   jtd=(int)(dtd/dt);
   nt=(data->nt-1)*jtd+1;
   base=0;
   redo=redo_src;
   create_transfer_sinc_data(8);

   encode=enc;
   rand_vec=r_vec;

   if(do_src && !redo){
     create_source_fields(1);
   }
}

void rtm_zero_op::create_transfer_sinc_data(int nsinc){
  jtd=(int)((data->dt)/dt);
  sinc_bob myr(jtd,nsinc);
  transfer_sinc_table_d(nsinc,jtd,myr.table);
}


bool rtm_zero_op::random_init(){
  //Define random encoding function
  int irand,count=0;
  int chk=1;
  bool fin=false;
  srand(time(NULL));
  int ns=data->nshots();
  while(fin==false){
    chk=0;
    for(int is=0; is<ns; is++){
      irand=rand()%2;
      if(irand==0) irand=-1;
      rand_vec[is]=irand;
      chk=chk+irand;
fprintf(stderr,"%d ",irand);

    }
    if(abs(chk)<=1) fin=true;
    count+=1;
  }
fprintf(stderr,"\n");
 return true;
}

bool rtm_zero_op::data_init(my_vector *bdvec){
  //Encode the data and output 

  data_rtm_3d *dn=(data_rtm_3d*) bdvec;
  bdata->set_source_file(dn);

  my_vector *dvec;
  data_rtm_3d *d=(data_rtm_3d*) dvec;
  data->set_source_file(d);

  bdvec->zero();

  int nt_big=8+data->nt;
  int npts=data->get_points();
  int npts1=data->get_points1();
  int npts2=data->get_points2();
  hypercube_float *src_p0=source->create_domain(1);

  float *s_z=new float[npts],*s_x=new float[npts],*s_y=new float[npts];
  axis ax_pt1(npts1);
  axis ax_pt2(npts2);

  time_t startdtimes, enddtimes;
  startdtimes = time(&startdtimes);

  float_3d *shot=new float_3d(data->get_axis(1),data->get_axis(2),data->get_axis(3));
  axis ax_t=axis(nt_big);
  data->get_source_func_encode(src_p0,0,rand_vec,s_z,s_x,s_y,8,nt_big,shot);

  bdata->add_data(0,shot);

  enddtimes = time(&enddtimes);

  fprintf(stderr,"      Data init, in %f \n",difftime(enddtimes, startdtimes));

  delete [] s_z;
  delete [] s_x;
  delete [] s_y;
  delete src_p0;
  delete shot;

  return true;

}

void rtm_zero_op::create_source_fields(int iter){
  //Propagate source wavefield through the random boundaries and save the final slices

  int nshotsx=data->nshotsx();
  int nshotsy=data->nshotsy();
  int nshots=data->nshots();
  auxinout("source_fields");

  int npts,nt_big;
  nt_big=9+data->nt; 
  npts=source->get_points(encode);
  int *locs=new int[npts];
  float *vals=new float[npts*(nt_big)];
  int *tempi=0;
  float *tempf=0;
  int seed=0;

  if(encode) nshots=1;

  for(int ishot=0; ishot< nshots; ishot++){

     hypercube_float *src_p0=source->create_domain(ishot);
     hypercube_float *src_p1=src_p0->clone();

     if(encode) source->get_source_func_encode(src_p0,ishot,nt_big,rand_vec,locs,vals);
     else source->get_source_func(src_p0,ishot,nt_big,locs,vals);

     transfer_source_func(npts,nt_big,locs,vals);

     //Seed our random boundary by both shot position and iteration number
     seed=locs[3] - ((int) locs[3]/10000)*10000;
     seed*=(iter);

     hypercube_float *vloc=vel->rand_subspace(seed,nboundt,nbound,nbound,dt,src_p0);

     transfer_vel_func1(nz,nx,ny,vloc->vals);

     float *temp=0;

     time_t startwtimes, endwtimes;
     startwtimes = time(&startwtimes);

     source_prop(nz,nx,ny,false,true,src_p0->vals,src_p1->vals,temp,jts,npts,nt,0,tempi,tempf,-1,"junk");

     endwtimes = time(&endwtimes);

     fprintf(stderr,"      Finished propagating wavefield, in %f \n",difftime(endwtimes, startwtimes));

     srite("source_fields",src_p0->vals,nz*nx*ny*4);
     srite("source_fields",src_p1->vals,nz*nx*ny*4);

     delete vloc; delete src_p0; delete src_p1;
   }

   delete [] locs;
   delete [] vals;
}

bool rtm_zero_op::forward(bool add, my_vector *mvec, my_vector *dvec, int iter){
  //Forward process

  image_rtm_3d *m=(image_rtm_3d*) mvec;
  image->set_source_file(m);

  data_rtm_3d *bd=(data_rtm_3d*) dvec;

  if(encode) bdata->set_source_file(bd);
  else data->set_source_file(bd);

  if(!add) dvec->zero();

  int npts_s=source->get_points(encode);

  int nt_big=8+data->nt;

  int npts=data->get_points();
  int npts1=data->get_points1();
  int npts2=data->get_points2();

  int *locs=new int[npts];
  int *locs_s=new int[npts_s];
  float *vals=new float[npts_s*(nt_big)];
  float *s_z=new float[npts],*s_x=new float[npts],*s_y=new float[npts];

  axis ax_pt1(npts1);
  axis ax_pt2(npts2);

  //Assume nz,nx,ny are the same for blended data and a single unblended shot
  float_3d *shot=new float_3d(data->get_axis(1),data->get_axis(2),data->get_axis(3));
  float_3d *shot_deriv=new float_3d(data->get_axis(1),data->get_axis(2),data->get_axis(3));


  axis ax_t=axis(nt_big);
  float_3d *rec_func=new float_3d(ax_t,ax_pt1,ax_pt2);

  int nshots=data->nshots();
  if(encode) nshots=1;

  int nts=(int)(nt/jts)+1;

  for(int ishot=0; ishot<nshots; ishot++){

    hypercube_float *src_p0=source->create_domain(ishot);
    hypercube_float *src_p1=src_p0->clone();
    axis ad1=src_p0->get_axis(1);
    axis ad2=src_p0->get_axis(2);
    axis ad3=src_p0->get_axis(3);

//    source->get_source_func_encode(src_p0,ishot,nt_big,rand_vec,locs_s,vals);
//    bdata->get_source_func(src_p0,ishot,s_z,s_x,s_y,8,nt_big,shot);
    if(encode){
      source->get_source_func_encode(src_p0,ishot,nt_big,rand_vec,locs_s,vals);
      bdata->get_source_func(src_p0,ishot,s_z,s_x,s_y,8,nt_big,shot);
    }
    else {
      source->get_source_func(src_p0,ishot,nt_big,locs_s,vals);
      data->get_source_func(src_p0,ishot,s_z,s_x,s_y,8,nt_big,shot);
    }

    transfer_source_func(npts_s,nt_big,locs_s,vals);
    int seed=locs_s[3] - ((int) locs_s[3]/10000)*10000;
    seed*=(iter);

    hypercube_float *vrand=vel->rand_subspace(seed,nboundt,nbound,nbound,dt,src_p0);
    hypercube_float *vnone=vel->rand_subspace(seed,0,0,0,dt,src_p0);
    hypercube_float *img=image->extract_sub(ad1,ad2,ad3);
    transfer_vel_func1(nz,nx,ny,vrand->vals);
    transfer_vel_func2(nz,nx,ny,vnone->vals);

    map_data mapit(npts,s_z,s_x,s_y,locs,src_p0,rec_func,shot_deriv,nt_big);
    transfer_receiver_func(npts,nt_big,locs,rec_func->vals);

    time_t startftimem, endftimem;
    startftimem = time(&startftimem);

    rtm_forward(ad1.n,ad2.n,ad3.n,jtd,img->vals,rec_func->vals,npts_s,npts,nt,nt_big,npts); 

    endftimem = time(&endftimem);
    fprintf(stderr,"      Forward, wall clock time = %f \n", difftime(endftimem, startftimem));

    deriv der=deriv(shot_deriv,shot);
    mapit.forward(false,rec_func,shot_deriv);
    der.forward(false,shot_deriv,shot);

    if(encode) bdata->add_data(0,shot);
    else data->add_data(ishot,shot);

    delete src_p0; delete src_p1; delete img;
    delete vrand; delete vnone;
  }
  delete [] s_z;
  delete [] s_x;
  delete [] s_y;
  delete [] locs_s;
  delete [] vals;
  delete [] locs;
  delete rec_func;
  delete shot;
  delete shot_deriv;

  return true;

}

bool rtm_zero_op::adjoint(bool add, my_vector *mvec, my_vector *dvec, int iter){
  //Adjoint process

  if(redo) create_source_fields(iter);
  if(!add) mvec->zero();

  data_rtm_3d *bd=(data_rtm_3d*) dvec;
  image_rtm_3d *m=(image_rtm_3d*) mvec;
  int npts_s=source->get_points(encode);

  if(encode) bdata->set_source_file(bd);
  else data->set_source_file(bd);

  image->set_source_file(m);
  int nt_big=8+data->nt;
  int npts=data->get_points();
  int npts1=data->get_points1();
  int npts2=data->get_points2();

  int *locs=new int[npts];
  int *locs_s=new int[npts_s];
  float *vals=new float[npts_s*(nt_big)];

  float *s_z=new float[npts],*s_x=new float[npts],*s_y=new float[npts];
  axis ax_pt1(npts1);
  axis ax_pt2(npts2);

  float_3d *shot=new float_3d(data->get_axis(1),data->get_axis(2),data->get_axis(3));
  float_3d *shot_deriv=new float_3d(data->get_axis(1),data->get_axis(2),data->get_axis(3));

  axis ax_t=axis(nt_big);

  float_3d *rec_func=new float_3d(ax_t,ax_pt1,ax_pt2);
  int seed=0;

  sseek("source_fields",0,0);
  image->zero();
  int nts=(int)(nt/jts)+1;

  int nshots=data->nshots();
  if(encode) nshots=1;

  for(int ishot=0; ishot<nshots; ishot++){
    hypercube_float *src_p0=source->create_domain(ishot);
    hypercube_float *src_p1=src_p0->clone();
    axis ad1=src_p0->get_axis(1);
    axis ad2=src_p0->get_axis(2);
    axis ad3=src_p0->get_axis(3);

    if(encode){
      source->get_source_func_encode(src_p0,ishot,nt_big,rand_vec,locs_s,vals);
      bdata->get_source_func(src_p0,ishot,s_z,s_x,s_y,8,nt_big,shot);
    }
    else {
      source->get_source_func(src_p0,ishot,nt_big,locs_s,vals);
      data->get_source_func(src_p0,ishot,s_z,s_x,s_y,8,nt_big,shot);
    }

    transfer_source_func(npts_s,nt_big,locs_s,vals);

    seed=locs_s[3] - ((int) locs_s[3]/10000)*10000;
    seed*=(iter);

    hypercube_float *vrand=vel->rand_subspace(seed,nboundt,nbound,nbound,dt,src_p0);
    hypercube_float *vnone=vel->rand_subspace(seed,0,0,0,dt,src_p0);
    hypercube_float *img=src_p0->clone();

    transfer_vel_func1(nz,nx,ny,vrand->vals);
    transfer_vel_func2(nz,nx,ny,vnone->vals);
    deriv der=deriv(shot_deriv,shot);
    der.adjoint(false,shot_deriv,shot);
    map_data mapit(npts,s_z,s_x,s_y,locs,src_p0,rec_func,shot_deriv,nt_big);
    mapit.adjoint(false,rec_func,shot_deriv);

    sreed("source_fields",src_p1->vals,4*ad1.n*ad2.n*ad3.n);
    sreed("source_fields",src_p0->vals,4*ad1.n*ad2.n*ad3.n);

    transfer_receiver_func(npts,nt_big,locs,rec_func->vals);

    float *temp2=new float[ad3.n];
    time_t startwtimem, endwtimem;
    startwtimem = time(&startwtimem);

    rtm_adjoint(ad1.n,ad2.n,ad3.n,jtd,src_p0->vals,src_p1->vals,img->vals,temp2,npts_s,npts,nt);

    endwtimem = time(&endwtimem);
    fprintf(stderr,"      Adjoint, wall clock time = %f \n", difftime(endwtimem, startwtimem));

    image->add_image(img);

    delete src_p0; delete src_p1; delete img;
    delete vrand; delete vnone;
  }
  sseek("source_fields",0,0);

  image->write_volume();


  delete [] s_z;
  delete [] s_x;
  delete [] s_y;
  delete [] locs_s;
  delete [] vals;
  delete [] locs;
  delete rec_func;
  delete shot;
  delete shot_deriv;

  return true;
}

