#include "source_func_3d.h"
#include<math.h>

source_func::source_func(char *tag){

   tag_init(tag);
   hypercube_float *tmp=new hypercube_float(this);
   wavelet=new hypercube_float(this);
   read_all(tag,tmp);
   wavelet->vals[get_axis(1).n-1]=0;

   //Define source function and take the first derivative
   for(int it=0; it < this->get_axis(1).n-1; it++) {
     wavelet->vals[it]=(tmp->vals[it+1]-tmp->vals[it])/tmp->get_axis(1).d;
   }

   delete tmp;
   dt=get_axis(1).d;

}

void source_func::set_sources_3d(float s_z, axis src_axis1, axis src_axis2){

  //Set up regular source axes
  for(int i2=0; i2 < src_axis2.n; i2++){
    for(int i1=0; i1 < src_axis1.n; i1++){
      sz.push_back(s_z);
      sx.push_back(src_axis1.o+src_axis1.d*i1);
      sy.push_back(src_axis2.o+src_axis2.d*i2);
    }
  }
  nshots=src_axis1.n*src_axis2.n;

}

int source_func::get_points(bool e ){
  int tmp;
  //If not encoding then do smeared source injection
  if(e) tmp=nshots;
  else tmp=9;
  return tmp;
}

void source_func::set_compute_size(hypercube *dom, float ap,int nbt, int nb, int nby, int fat,
  int blocksize){
  az=dom->get_axis(1);
  ax=dom->get_axis(2);
  ay=dom->get_axis(3); 

  nbound=nb+fat; nboundt=nbt; nbound_y=nby+fat;
  aper=ap;

  ay.n+=2*nbound_y;
  ax.n+=2*nbound;
  az.n+=nbt+nb+2*fat;

  //Pad domain to fit into an even number of GPU blocks. Currently using a static velocity model however the commented section below allows for model windowing according to a defined aperture

  int rem_x=ax.n-((int) (ax.n/16))*16;
  int rem_y=ay.n-((int) (ay.n/16))*16;
  int rem_z=az.n-((int) (az.n/16))*16;

  if(rem_y!=0) rem_y=16-rem_y;
  if(rem_z!=0) rem_z=16-rem_z;
  if(rem_x!=0) rem_x=16-rem_x;

  fprintf(stderr,"  Remainders are %d %d %d from %d %d %d. Padding z with %d + %d ; %d\n",rem_z,rem_x,rem_y,az.n,ax.n,ay.n,nbt,nb,nboundt);

  ay.n+=rem_y;
  ax.n+=rem_x;
  az.n+=rem_z;

 /* ay.n=(int)(ap/ay.d)*2+1+2*nbound_y;
  ax.n=(int)(ap/ax.d)*2+1+2*nbound;
  az.n=az.n+nbt+nb+fat*2;

  int t1=(ax.n-fat*2)/blocksize;
  int rem=ax.n-t1*blocksize-2*fat;
  if(rem!=0) rem=blocksize-rem;
  if(rem!=0) ax.n+=+rem;

  t1=(ay.n-fat*2)/blocksize;
  rem=ay.n-t1*blocksize-2*fat;
  if(rem!=0) rem=blocksize-rem;
  if(rem!=0) ay.n+=+rem;
    
  t1=(az.n-fat*2)/blocksize;
  rem=az.n-t1*blocksize-2*fat;
  if(rem!=0) rem=blocksize-rem;
  if(rem!=0) az.n+=+rem;
*/

//  az.o=-20.;//az.o;//+(nbt+fat)*az.d;
}

hypercube_float *source_func::create_domain(int ishot){
  //Define the domain for the source wavefield slice
  az.o=-az.d*nboundt;
  ax.o=-ax.d*nbound;
  ay.o=-ay.d*nbound;
  std::vector<axis>  axes; axes.push_back(az); axes.push_back(ax); axes.push_back(ay);
  hypercube_float *tmp=new hypercube_float(axes);
  return tmp; 
}

void source_func::get_source_func(hypercube_float *domain, int ishot, int nts, int *locs, float *time){

  //Create source function, non encoded and smeared about z and x axes
  axis a1=domain->get_axis(1);
  axis a2=domain->get_axis(2);
  axis a3=domain->get_axis(3);
  axis at=get_axis(1);

  float fy=(sy[ishot]-a3.o)/a3.d;
  float fx=(sx[ishot]-a2.o)/a2.d;
  float fz=(sz[ishot]-a1.o)/a1.d;
  int iy=(int)fy;
  int ix=(int)fx;
  int iz=(int)fz;
  int i=0;
fprintf(stderr,"  Source is at %d %d %d ; nts=%d\n",iz,ix,iy,nts);

  for(int it=0; it < nts*9; it++){
    time[it]=0;
  }
  for(int i3=iy-1; i3 <= iy-1; i3++){
    float yn=fy-i3;
    for(int i2=ix-1; i2 <= ix+1; i2++){
      float xn=fx-i2;
      for(int i1=iz-1; i1<= iz+1; i1++,i++){
        float zn=fz-i1;
        locs[i]=i1+i2*a1.n+i3*a1.n*a2.n;
        float scale=expf(-zn*zn-xn*xn-yn*yn);
        for(int it=0; it < at.n; it++){
          time[nts*i+it+4]=scale*wavelet->vals[it];
        }
      }
    }
  }

}

void source_func::get_source_func_encode(hypercube_float *domain, int ishot, int nts , int *rvec, int *locs, float *time){

  //Set up encoded source function according to rvec
  axis a1=domain->get_axis(1);
  axis a2=domain->get_axis(2);
  axis a3=domain->get_axis(3);
  axis at=get_axis(1);

  for(int it=0; it < nts*nshots; it++){
    time[it]=0;
  }

  float fy=0;
  float fx=0;
  float fz=0;
  int iy=0;
  int ix=0;
  int iz=0;
  for(int is=0; is<nshots; is++){
    fy=(sy[is]-a3.o)/a3.d;
    fx=(sx[is]-a2.o)/a2.d;
    fz=(sz[is]-a1.o)/a1.d;
    iy=(int)fy;
    ix=(int)fx;
    iz=(int)fz;
    locs[is]=iz+ix*a1.n+iy*a1.n*a2.n;
    for(int it=0; it < at.n; it++){
      time[nts*is+it+4]=rvec[is]*wavelet->vals[it];
    }
  }

}

