#define BLOCKZ_SIZE 16
#define BLOCKX_SIZE 16
#define BLOCKY_SIZE 16
#define FAT 7
#define COEFFS_SIZE 22

#define C0  0
#define CZ1 1
#define CX1 2
#define CY1 3
#define CZ2 4
#define CX2 5
#define CY2 6
#define CZ3 7
#define CX3 8
#define CY3 9
#define CZ4 10
#define CX4 11
#define CY4 12
#define CZ5 13
#define CX5 14
#define CY5 15
#define CZ6 16
#define CX6 17
#define CY6 18
#define CZ7 19
#define CX7 20
#define CY7 21

#define C_C00(d) (7.0/(4.0*(d)*(d)))
#define get_coeffs(d1,d2,d3) {-952.0/551.0*(C_C00(d1)+C_C00(d2)+C_C00(d3)), C_C00(d1), C_C00(d2), C_C00(d3),-C_C00(d1)/6.0,-C_C00(d2)/6.0,-C_C00(d3)/6.0,C_C00(d1)/27.0,C_C00(d2)/27.0,C_C00(d3)/27.0,-C_C00(d1)/132.0,-C_C00(d2)/132.0,-C_C00(d3)/132.0,C_C00(d1)/825.0,C_C00(d2)/825.0,C_C00(d3)/825.0,-C_C00(d1)/7722.0,-C_C00(d2)/7722.0,-C_C00(d3)/7722.0,C_C00(d1)/147147.0,C_C00(d2)/147147.0,C_C00(d3)/147147.0}
__constant__ float coeffs[COEFFS_SIZE];

//first off, going to try and use 3D texture
extern "C" __global__ void wave_kernel(float *p0, float *p1, float *p2){
	__shared__ float p1s[BLOCKX_SIZE+2*FAT][BLOCKY_SIZE+2*FAT];

	// define coordinates for the slow axis - everything else will follow from here
	int nblocksz=(n1gpu-2*FAT)/BLOCKZ_SIZE;
	int yblock=(blockIdx.x/nblocksz);
	int y0=FAT+yblock*BLOCKY_SIZE;

	int il=FAT+threadIdx.x; 				// thread-local z coordinate
	int jl=FAT+threadIdx.y;					// thread-local x coordinate

	int ig=(blockIdx.x-yblock*nblocksz)*blockDim.x+il; 	// global z coordinates
	int jg=blockIdx.y*blockDim.y+jl; 	 		// global x coordinates

	float p1y_p1, p1y_m1, p1y_p2, p1y_m2, p1y_p3, p1y_m3, p1y_p4, p1y_m4;
	float p1y_p5, p1y_m5, p1y_p6, p1y_m6, p1y_p7, p1y_m7;

	int stride=n1gpu*n2gpu; //this corresponds to distance in elements between wavefield slices
	int addr=ig+n1gpu*jg+stride*y0;
	int addr_fwd=addr-7*stride;//-stride-stride-stride-stride-stride-stride;

	p1y_m6=p1[addr_fwd];
	p1y_m5=p1[addr_fwd+=stride];
	p1y_m4=p1[addr_fwd+=stride];
	p1y_m3=p1[addr_fwd+=stride];
	p1y_m2=p1[addr_fwd+=stride];
	p1y_m1=p1[addr_fwd+=stride];
	p1s[jl][il]=p1[addr_fwd+=stride];
	p1y_p1=p1[addr_fwd+=stride];
	p1y_p2=p1[addr_fwd+=stride];
	p1y_p3=p1[addr_fwd+=stride];
	p1y_p4=p1[addr_fwd+=stride];
	p1y_p5=p1[addr_fwd+=stride];
	p1y_p6=p1[addr_fwd+=stride];
	p1y_p7=p1[addr_fwd+=stride];

	// loop over slow axis. ++yl means variable is used before incremented
	for(int yl=0; yl<BLOCKY_SIZE; ++yl){

		int y=y0+yl;

		//update slow axis values
		p1y_m7=p1y_m6;		
		p1y_m6=p1y_m5;		
		p1y_m5=p1y_m4;
		p1y_m4=p1y_m3;		
		p1y_m3=p1y_m2;		
		p1y_m2=p1y_m1;
		p1y_m1=p1s[jl][il];	
		p1s[jl][il]=p1y_p1;
		p1y_p1=p1y_p2;		
		p1y_p2=p1y_p3;		
		p1y_p3=p1y_p4;
		p1y_p4=p1y_p5;		
		p1y_p5=p1y_p6;		
		p1y_p6=p1y_p7;
		p1y_p7=p1[addr_fwd+=stride];

    
		if(threadIdx.x<FAT){
			p1s[jl][threadIdx.x]=p1[addr-FAT];
			p1s[jl][il+BLOCKZ_SIZE]=p1[addr+BLOCKZ_SIZE];
		}
		if(threadIdx.y<FAT){
			p1s[threadIdx.y][il]=p1[addr-FAT*n1gpu];
			p1s[jl+BLOCKX_SIZE][il]=p1[addr+BLOCKX_SIZE*n1gpu];
		}
		__syncthreads();

		p2[addr]=
		tex3D(vel_ref1,ig,jg,y)*
			(
			 coeffs[C0]*p1s[jl][il]
			+coeffs[CZ1]*(p1s[jl][il+1]+p1s[jl][il-1])
			+coeffs[CX1]*(p1s[jl+1][il]+p1s[jl-1][il])
			+coeffs[CY1]*(p1y_p1+p1y_m1)
			+coeffs[CZ2]*(p1s[jl][il+2]+p1s[jl][il-2])
			+coeffs[CX2]*(p1s[jl+2][il]+p1s[jl-2][il])
			+coeffs[CY2]*(p1y_p2+p1y_m2)
			+coeffs[CZ3]*(p1s[jl][il+3]+p1s[jl][il-3])
			+coeffs[CX3]*(p1s[jl+3][il]+p1s[jl-3][il])
			+coeffs[CY3]*(p1y_p3+p1y_m3)
			+coeffs[CZ4]*(p1s[jl][il+4]+p1s[jl][il-4])
			+coeffs[CX4]*(p1s[jl+4][il]+p1s[jl-4][il])
			+coeffs[CY4]*(p1y_p4+p1y_m4)
			+coeffs[CZ5]*(p1s[jl][il+5]+p1s[jl][il-5])
			+coeffs[CX5]*(p1s[jl+5][il]+p1s[jl-5][il])
			+coeffs[CY5]*(p1y_p5+p1y_m5)
			+coeffs[CZ6]*(p1s[jl][il+6]+p1s[jl][il-6])
			+coeffs[CX6]*(p1s[jl+6][il]+p1s[jl-6][il])
			+coeffs[CY6]*(p1y_p6+p1y_m6)
			+coeffs[CZ7]*(p1s[jl][il+7]+p1s[jl][il-7])
			+coeffs[CX7]*(p1s[jl+7][il]+p1s[jl-7][il])
			+coeffs[CY7]*(p1y_p7+p1y_m7)
			)
			+p1s[jl][il]+p1s[jl][il]-p0[addr];
			addr+=stride;
	}
}

//same as wave_kernel but uses vel_ref2 (non-random boundary velocity)


extern "C" __global__ void wave_kernel2(float *p0, float *p1, float *p2){
	__shared__ float p1s[BLOCKX_SIZE+2*FAT][BLOCKY_SIZE+2*FAT];

	// define coordinates for the slow axis - everything else will follow from here
	int nblocksz=(n1gpu-2*FAT)/BLOCKZ_SIZE;
	int yblock=(blockIdx.x/nblocksz);
	int y0=FAT+yblock*BLOCKY_SIZE;

	int il=FAT+threadIdx.x; 				// thread-local z coordinate
	int jl=FAT+threadIdx.y;					// thread-local x coordinate

	int ig=(blockIdx.x-yblock*nblocksz)*blockDim.x+il; 	// global z coordinates
	int jg=blockIdx.y*blockDim.y+jl; 	 		// global x coordinates

	float 	p1y_p1, p1y_m1, p1y_p2, p1y_m2, p1y_p3, p1y_m3, p1y_p4, p1y_m4, p1y_p5, p1y_m5, p1y_p6, p1y_m6, p1y_p7, p1y_m7; //here we define the y axis values

	int stride=n1gpu*n2gpu; //this corresponds to distance in elements between wavefield slices
	int addr=ig+n1gpu*jg+stride*y0;
	int addr_fwd=addr-stride-stride-stride-stride-stride-stride-stride; //these (supposedly!) give the memorry location for the current point
		//note that Paulius has a slightly different term here - check one day!

	//copy to shared memory
	p1y_m6=p1[addr_fwd];
	p1y_m5=p1[addr_fwd+=stride];
	p1y_m4=p1[addr_fwd+=stride];
	p1y_m3=p1[addr_fwd+=stride];
	p1y_m2=p1[addr_fwd+=stride];
	p1y_m1=p1[addr_fwd+=stride];
	p1s[jl][il]=p1[addr_fwd+=stride];
	p1y_p1=p1[addr_fwd+=stride];
	p1y_p2=p1[addr_fwd+=stride];
	p1y_p3=p1[addr_fwd+=stride];
	p1y_p4=p1[addr_fwd+=stride];
	p1y_p5=p1[addr_fwd+=stride];
	p1y_p6=p1[addr_fwd+=stride];
	p1y_p7=p1[addr_fwd+=stride];

	// loop over slow axis. ++yl means variable is used before incremented
	for(int yl=0; yl<BLOCKY_SIZE; ++yl){

		int y=y0+yl;

		//update slow axis values
		p1y_m7=p1y_m6;		
		p1y_m6=p1y_m5;		
		p1y_m5=p1y_m4;
		p1y_m4=p1y_m3;		
		p1y_m3=p1y_m2;		
		p1y_m2=p1y_m1;
		p1y_m1=p1s[jl][il];	
		p1s[jl][il]=p1y_p1;
		p1y_p1=p1y_p2;		
		p1y_p2=p1y_p3;		
		p1y_p3=p1y_p4;
		p1y_p4=p1y_p5;		
		p1y_p5=p1y_p6;		
		p1y_p6=p1y_p7;
		p1y_p7=p1[addr_fwd+=stride];

    
		if(threadIdx.x<FAT){
			p1s[jl][threadIdx.x]=p1[addr-FAT];
			p1s[jl][il+BLOCKZ_SIZE]=p1[addr+BLOCKZ_SIZE];
		}
		if(threadIdx.y<FAT){
			p1s[threadIdx.y][il]=p1[addr-FAT*n1gpu];
			p1s[jl+BLOCKX_SIZE][il]=p1[addr+BLOCKX_SIZE*n1gpu];
		}
		__syncthreads();

		p2[addr]=//10;
		tex3D(vel_ref2,ig,jg,y)*
			(
			 coeffs[C0]*p1s[jl][il]
			+coeffs[CZ1]*(p1s[jl][il+1]+p1s[jl][il-1])
			+coeffs[CX1]*(p1s[jl+1][il]+p1s[jl-1][il])
			+coeffs[CY1]*(p1y_p1+p1y_m1)
			+coeffs[CZ2]*(p1s[jl][il+2]+p1s[jl][il-2])
			+coeffs[CX2]*(p1s[jl+2][il]+p1s[jl-2][il])
			+coeffs[CY2]*(p1y_p2+p1y_m2)
			+coeffs[CZ3]*(p1s[jl][il+3]+p1s[jl][il-3])
			+coeffs[CX3]*(p1s[jl+3][il]+p1s[jl-3][il])
			+coeffs[CY3]*(p1y_p3+p1y_m3)
			+coeffs[CZ4]*(p1s[jl][il+4]+p1s[jl][il-4])
			+coeffs[CX4]*(p1s[jl+4][il]+p1s[jl-4][il])
			+coeffs[CY4]*(p1y_p4+p1y_m4)
			+coeffs[CZ5]*(p1s[jl][il+5]+p1s[jl][il-5])
			+coeffs[CX5]*(p1s[jl+5][il]+p1s[jl-5][il])
			+coeffs[CY5]*(p1y_p5+p1y_m5)
			+coeffs[CZ6]*(p1s[jl][il+6]+p1s[jl][il-6])
			+coeffs[CX6]*(p1s[jl+6][il]+p1s[jl-6][il])
			+coeffs[CY6]*(p1y_p6+p1y_m6)
			+coeffs[CZ7]*(p1s[jl][il+7]+p1s[jl][il-7])
			+coeffs[CX7]*(p1s[jl+7][il]+p1s[jl-7][il])
			+coeffs[CY7]*(p1y_p7+p1y_m7)
			)
			+p1s[jl][il]+p1s[jl][il]-p0[addr];
			addr+=stride;
	}
}

extern "C" __global__ void damp_kernel(int nblocksz, float *p0, float *p1){
	int yblock=(blockIdx.x/nblocksz);
	int y0=FAT+yblock*BLOCKY_SIZE;
	int zl=FAT+threadIdx.x; 				// local  zdir
	int zg=(blockIdx.x-yblock*nblocksz)*blockDim.x+zl; 	// global zdir
	int xl=FAT+threadIdx.y; 				// local  xdir
	int xg=blockIdx.y*blockDim.y+xl; 			// global xdir	

	int stride=n1gpu*n2gpu;
	int bc_agpu=50;
	float bc_bgpu=0.0005;

	for(int yl=0; yl<BLOCKY_SIZE; ++yl){
		int y=y0+yl;
		int edge=min2(zg-FAT , min2(y-FAT,min2(min2(n2gpu-FAT-xg,n1gpu-FAT-zg),min2(n3gpu-FAT-y,xg-FAT))));
		if(edge>=0){
		  int addr=zg+n1gpu*xg+stride*y;
		  float temp=expf(-bc_bgpu*(bc_agpu-edge));
		  if(temp<1.){
		    p1[addr]*=temp;
		    p0[addr]*=temp;
		  }
		}
	}
}

extern "C" __global__ void damp_top_kernel(int nblocksz, float *p0, float *p1){
	int yblock=(blockIdx.x/nblocksz);
	int y0=FAT+yblock*BLOCKY_SIZE;
	int zl=FAT+threadIdx.x; 				// local  zdir
	int zg=(blockIdx.x-yblock*nblocksz)*blockDim.x+zl; 	// global zdir
	int xl=FAT+threadIdx.y; 				// local  xdir
	int xg=blockIdx.y*blockDim.y+xl; 			// global xdir	

	int stride=n1gpu*n2gpu;
	int bc_agpu=50;
	float bc_bgpu=0.0005;

	for(int yl=0; yl<BLOCKY_SIZE; ++yl){
		int y=y0+yl;
		int edge=n1gpu-FAT-zg;//min2(zg-FAT ,n1gpu-FAT-zg);
		if(edge>=0){
		  int addr=zg+n1gpu*xg+stride*y;
		  float temp=expf(-bc_bgpu*(bc_agpu-edge));
		  if(temp<1.){
		    p1[addr]*=temp;
		    p0[addr]*=temp;
		  }
		}
	}
}

extern "C" __global__ void wave_kernel_adj(float *p0, float *p1, float *p2){
	__shared__ float p1s[BLOCKX_SIZE+2*FAT][BLOCKY_SIZE+2*FAT];
	__shared__ float vls[BLOCKX_SIZE+2*FAT][BLOCKY_SIZE+2*FAT];

	// define coordinates for the slow axis - everything else will follow from here
	int nblocksz=(n1gpu-2*FAT)/BLOCKZ_SIZE;
	int yblock=(blockIdx.x/nblocksz);
	int y0=FAT+yblock*BLOCKY_SIZE;

	int il=FAT+threadIdx.x; 				// thread-local z coordinate
	int jl=FAT+threadIdx.y;					// thread-local x coordinate

	int ig=(blockIdx.x-yblock*nblocksz)*blockDim.x+il; 	// global z coordinates
	int jg=blockIdx.y*blockDim.y+jl; 	 		// global x coordinates

	float 	p1y_p1, p1y_m1, p1y_p2, p1y_m2, p1y_p3, p1y_m3, p1y_p4, p1y_m4, p1y_p5, p1y_m5, p1y_p6, p1y_m6, p1y_p7, p1y_m7;
	float 	vy_p1, vy_m1, vy_p2, vy_m2, vy_p3, vy_m3, vy_p4, vy_m4, vy_p5, vy_m5, vy_p6, vy_m6, vy_p7, vy_m7; 

	int stride=n1gpu*n2gpu; //this corresponds to distance in elements between wavefield slices
	int addr=ig+n1gpu*jg+stride*y0;
	int addr_fwd=addr-stride-stride-stride-stride-stride-stride-stride; //these (supposedly!) give the memorry location for the current point
		//note that Paulius has a slightly different term here - check one day!

	//copy to shared memory
	p1y_m6=p1[addr_fwd];
	p1y_m5=p1[addr_fwd+=stride];
	p1y_m4=p1[addr_fwd+=stride];
	p1y_m3=p1[addr_fwd+=stride];
	p1y_m2=p1[addr_fwd+=stride];
	p1y_m1=p1[addr_fwd+=stride];
	p1s[jl][il]=p1[addr_fwd+=stride];
	p1y_p1=p1[addr_fwd+=stride];
	p1y_p2=p1[addr_fwd+=stride];
	p1y_p3=p1[addr_fwd+=stride];
	p1y_p4=p1[addr_fwd+=stride];
	p1y_p5=p1[addr_fwd+=stride];
	p1y_p6=p1[addr_fwd+=stride];
	p1y_p7=p1[addr_fwd+=stride];

	vy_m7=tex3D(vel_ref1,ig,jg,y0-7);
	vy_m6=tex3D(vel_ref1,ig,jg,y0-7);
	vy_m5=tex3D(vel_ref1,ig,jg,y0-6);
	vy_m4=tex3D(vel_ref1,ig,jg,y0-5);
	vy_m3=tex3D(vel_ref1,ig,jg,y0-4);
	vy_m2=tex3D(vel_ref1,ig,jg,y0-3);
	vy_m1=tex3D(vel_ref1,ig,jg,y0-2);
	vls[jl][il]=tex3D(vel_ref1,ig,jg,y0-1);
	vy_p1=tex3D(vel_ref1,ig,jg,y0);
	vy_p2=tex3D(vel_ref1,ig,jg,y0+1);
	vy_p3=tex3D(vel_ref1,ig,jg,y0+2);
	vy_p4=tex3D(vel_ref1,ig,jg,y0+3);
	vy_p5=tex3D(vel_ref1,ig,jg,y0+4);
	vy_p6=tex3D(vel_ref1,ig,jg,y0+5);
	vy_p7=tex3D(vel_ref1,ig,jg,y0+6);

	// loop over slow axis. ++yl means variable is used before incremented
	for(int yl=0; yl<BLOCKY_SIZE; ++yl){

		int y=y0+yl;

		//update slow axis values
		p1y_m7=p1y_m6;		
		p1y_m6=p1y_m5;		
		p1y_m5=p1y_m4;
		p1y_m4=p1y_m3;		
		p1y_m3=p1y_m2;		
		p1y_m2=p1y_m1;
		p1y_m1=p1s[jl][il];	
		p1s[jl][il]=p1y_p1;
		p1y_p1=p1y_p2;		
		p1y_p2=p1y_p3;		
		p1y_p3=p1y_p4;
		p1y_p4=p1y_p5;		
		p1y_p5=p1y_p6;		
		p1y_p6=p1y_p7;
		p1y_p7=p1[addr_fwd+=stride];

		vy_m7=vy_m6;		
		vy_m6=vy_m5;		
		vy_m5=vy_m4;
		vy_m4=vy_m3;		
		vy_m3=vy_m2;		
		vy_m2=vy_m1;
		vy_m1=vls[jl][il];	
		vls[jl][il]=vy_p1;
		vy_p1=vy_p2;		
		vy_p2=vy_p3;		
		vy_p3=vy_p4;
		vy_p4=vy_p5;		
		vy_p5=vy_p6;		
		vy_p6=vy_p7;
		vy_p7=tex3D(vel_ref1,ig,jg,y+7);

		if(threadIdx.x<FAT){
			p1s[jl][threadIdx.x]=p1[addr-FAT];
			p1s[jl][il+BLOCKZ_SIZE]=p1[addr+BLOCKZ_SIZE];
			vls[jl][threadIdx.x]=tex3D(vel_ref1,ig-FAT,jg,y);
			vls[jl][il+BLOCKZ_SIZE]=tex3D(vel_ref1,ig+BLOCKZ_SIZE,jg,y);
		}
		if(threadIdx.y<FAT){
			p1s[threadIdx.y][il]=p1[addr-FAT*n1gpu];
			p1s[jl+BLOCKX_SIZE][il]=p1[addr+BLOCKX_SIZE*n1gpu];
			vls[threadIdx.y][il]=tex3D(vel_ref1,ig,jg-FAT,y);
			vls[jl+BLOCKX_SIZE][il]=tex3D(vel_ref1,ig,jg+BLOCKX_SIZE,y);
		}
		__syncthreads();

		p2[addr]=
			 coeffs[C0]*p1s[jl][il]*vls[jl][il]
			+coeffs[CZ1]*(p1s[jl][il+1]*vls[jl][il+1]+p1s[jl][il-1]*vls[jl][il-1])
			+coeffs[CX1]*(p1s[jl+1][il]*vls[jl+1][il]+p1s[jl-1][il]*vls[jl-1][il])
			+coeffs[CY1]*(p1y_p1*vy_p1+p1y_m1*vy_m1)
			+coeffs[CZ2]*(p1s[jl][il+2]*vls[jl][il+2]+p1s[jl][il-2]*vls[jl][il-2])
			+coeffs[CX2]*(p1s[jl+2][il]*vls[jl+2][il]+p1s[jl-2][il]*vls[jl-2][il])
			+coeffs[CY2]*(p1y_p2*vy_p2+p1y_m2*vy_m2)
			+coeffs[CZ3]*(p1s[jl][il+3]*vls[jl][il+3]+p1s[jl][il-3]*vls[jl][il-3])
			+coeffs[CX3]*(p1s[jl+3][il]*vls[jl+3][il]+p1s[jl-3][il]*vls[jl-3][il])
			+coeffs[CY3]*(p1y_p3*vy_p3+p1y_m3*vy_m3)
			+coeffs[CZ4]*(p1s[jl][il+4]*vls[jl][il+4]+p1s[jl][il-4]*vls[jl][il-4])
			+coeffs[CX4]*(p1s[jl+4][il]*vls[jl+4][il]+p1s[jl-4][il]*vls[jl-4][il])
			+coeffs[CY4]*(p1y_p4*vy_p4+p1y_m4*vy_m4)
			+coeffs[CZ5]*(p1s[jl][il+5]*vls[jl][il+5]+p1s[jl][il-5]*vls[jl][il-5])
			+coeffs[CX5]*(p1s[jl+5][il]*vls[jl+5][il]+p1s[jl-5][il]*vls[jl-5][il])
			+coeffs[CY5]*(p1y_p5*vy_p5+p1y_m5*vy_m5)
			+coeffs[CZ6]*(p1s[jl][il+6]*vls[jl][il+6]+p1s[jl][il-6]*vls[jl][il-6])
			+coeffs[CX6]*(p1s[jl+6][il]*vls[jl+6][il]+p1s[jl-6][il]*vls[jl-6][il])
			+coeffs[CY6]*(p1y_p6*vy_p6+p1y_m6*vy_m6)
			+coeffs[CZ7]*(p1s[jl][il+7]*vls[jl][il+7]+p1s[jl][il-7]*vls[jl][il-7])
			+coeffs[CX7]*(p1s[jl+7][il]*vls[jl+7][il]+p1s[jl-7][il]*vls[jl-7][il])
			+coeffs[CY7]*(p1y_p7*vy_p7+p1y_m7*vy_m7)
			+p1s[jl][il]+p1s[jl][il]-p0[addr];
		addr+=stride;
	}

}

extern "C" __global__ void wave_kernel_adj2(float *p0, float *p1, float *p2){
	__shared__ float p1s[BLOCKX_SIZE+2*FAT][BLOCKY_SIZE+2*FAT];
	__shared__ float vls[BLOCKX_SIZE+2*FAT][BLOCKY_SIZE+2*FAT];

	// define coordinates for the slow axis - everything else will follow from here
	int nblocksz=(n1gpu-2*FAT)/BLOCKZ_SIZE;
	int yblock=(blockIdx.x/nblocksz);
	int y0=FAT+yblock*BLOCKY_SIZE;

	int il=FAT+threadIdx.x; 				// thread-local z coordinate
	int jl=FAT+threadIdx.y;					// thread-local x coordinate

	int ig=(blockIdx.x-yblock*nblocksz)*blockDim.x+il; 	// global z coordinates
	int jg=blockIdx.y*blockDim.y+jl; 	 		// global x coordinates

	float 	p1y_p1, p1y_m1, p1y_p2, p1y_m2, p1y_p3, p1y_m3, p1y_p4, p1y_m4, p1y_p5, p1y_m5, p1y_p6, p1y_m6, p1y_p7, p1y_m7;
	float 	vy_p1, vy_m1, vy_p2, vy_m2, vy_p3, vy_m3, vy_p4, vy_m4, vy_p5, vy_m5, vy_p6, vy_m6, vy_p7, vy_m7; 

	int stride=n1gpu*n2gpu; //this corresponds to distance in elements between wavefield slices
	int addr=ig+n1gpu*jg+stride*y0;
	int addr_fwd=addr-stride-stride-stride-stride-stride-stride-stride; //these (supposedly!) give the memorry location for the current point
		//note that Paulius has a slightly different term here - check one day!

	//copy to shared memory
	p1y_m6=p1[addr_fwd];
	p1y_m5=p1[addr_fwd+=stride];
	p1y_m4=p1[addr_fwd+=stride];
	p1y_m3=p1[addr_fwd+=stride];
	p1y_m2=p1[addr_fwd+=stride];
	p1y_m1=p1[addr_fwd+=stride];
	p1s[jl][il]=p1[addr_fwd+=stride];
	p1y_p1=p1[addr_fwd+=stride];
	p1y_p2=p1[addr_fwd+=stride];
	p1y_p3=p1[addr_fwd+=stride];
	p1y_p4=p1[addr_fwd+=stride];
	p1y_p5=p1[addr_fwd+=stride];
	p1y_p6=p1[addr_fwd+=stride];
	p1y_p7=p1[addr_fwd+=stride];

	vy_m7=tex3D(vel_ref2,ig,jg,y0-7);
	vy_m6=tex3D(vel_ref2,ig,jg,y0-7);
	vy_m5=tex3D(vel_ref2,ig,jg,y0-6);
	vy_m4=tex3D(vel_ref2,ig,jg,y0-5);
	vy_m3=tex3D(vel_ref2,ig,jg,y0-4);
	vy_m2=tex3D(vel_ref2,ig,jg,y0-3);
	vy_m1=tex3D(vel_ref2,ig,jg,y0-2);
	vls[jl][il]=tex3D(vel_ref2,ig,jg,y0-1);
	vy_p1=tex3D(vel_ref2,ig,jg,y0);
	vy_p2=tex3D(vel_ref2,ig,jg,y0+1);
	vy_p3=tex3D(vel_ref2,ig,jg,y0+2);
	vy_p4=tex3D(vel_ref2,ig,jg,y0+3);
	vy_p5=tex3D(vel_ref2,ig,jg,y0+4);
	vy_p6=tex3D(vel_ref2,ig,jg,y0+5);
	vy_p7=tex3D(vel_ref2,ig,jg,y0+6);

	// loop over slow axis. ++yl means variable is used before incremented
	for(int yl=0; yl<BLOCKY_SIZE; ++yl){

		int y=y0+yl;

		//update slow axis values
		p1y_m7=p1y_m6;		
		p1y_m6=p1y_m5;		
		p1y_m5=p1y_m4;
		p1y_m4=p1y_m3;		
		p1y_m3=p1y_m2;		
		p1y_m2=p1y_m1;
		p1y_m1=p1s[jl][il];	
		p1s[jl][il]=p1y_p1;
		p1y_p1=p1y_p2;		
		p1y_p2=p1y_p3;		
		p1y_p3=p1y_p4;
		p1y_p4=p1y_p5;		
		p1y_p5=p1y_p6;		
		p1y_p6=p1y_p7;
		p1y_p7=p1[addr_fwd+=stride];

		vy_m7=vy_m6;		
		vy_m6=vy_m5;		
		vy_m5=vy_m4;
		vy_m4=vy_m3;		
		vy_m3=vy_m2;		
		vy_m2=vy_m1;
		vy_m1=vls[jl][il];	
		vls[jl][il]=vy_p1;
		vy_p1=vy_p2;		
		vy_p2=vy_p3;		
		vy_p3=vy_p4;
		vy_p4=vy_p5;		
		vy_p5=vy_p6;		
		vy_p6=vy_p7;
		vy_p7=tex3D(vel_ref2,ig,jg,y+7);

		if(threadIdx.x<FAT){
			p1s[jl][threadIdx.x]=p1[addr-FAT];
			p1s[jl][il+BLOCKZ_SIZE]=p1[addr+BLOCKZ_SIZE];
			vls[jl][threadIdx.x]=tex3D(vel_ref2,ig-FAT,jg,y);
			vls[jl][il+BLOCKZ_SIZE]=tex3D(vel_ref2,ig+BLOCKZ_SIZE,jg,y);
		}
		if(threadIdx.y<FAT){
			p1s[threadIdx.y][il]=p1[addr-FAT*n1gpu];
			p1s[jl+BLOCKX_SIZE][il]=p1[addr+BLOCKX_SIZE*n1gpu];
			vls[threadIdx.y][il]=tex3D(vel_ref2,ig,jg-FAT,y);
			vls[jl+BLOCKX_SIZE][il]=tex3D(vel_ref2,ig,jg+BLOCKX_SIZE,y);
		}
		__syncthreads();

		p2[addr]=
			 coeffs[C0]*p1s[jl][il]*vls[jl][il]
			+coeffs[CZ1]*(p1s[jl][il+1]*vls[jl][il+1]+p1s[jl][il-1]*vls[jl][il-1])
			+coeffs[CX1]*(p1s[jl+1][il]*vls[jl+1][il]+p1s[jl-1][il]*vls[jl-1][il])
			+coeffs[CY1]*(p1y_p1*vy_p1+p1y_m1*vy_m1)
			+coeffs[CZ2]*(p1s[jl][il+2]*vls[jl][il+2]+p1s[jl][il-2]*vls[jl][il-2])
			+coeffs[CX2]*(p1s[jl+2][il]*vls[jl+2][il]+p1s[jl-2][il]*vls[jl-2][il])
			+coeffs[CY2]*(p1y_p2*vy_p2+p1y_m2*vy_m2)
			+coeffs[CZ3]*(p1s[jl][il+3]*vls[jl][il+3]+p1s[jl][il-3]*vls[jl][il-3])
			+coeffs[CX3]*(p1s[jl+3][il]*vls[jl+3][il]+p1s[jl-3][il]*vls[jl-3][il])
			+coeffs[CY3]*(p1y_p3*vy_p3+p1y_m3*vy_m3)
			+coeffs[CZ4]*(p1s[jl][il+4]*vls[jl][il+4]+p1s[jl][il-4]*vls[jl][il-4])
			+coeffs[CX4]*(p1s[jl+4][il]*vls[jl+4][il]+p1s[jl-4][il]*vls[jl-4][il])
			+coeffs[CY4]*(p1y_p4*vy_p4+p1y_m4*vy_m4)
			+coeffs[CZ5]*(p1s[jl][il+5]*vls[jl][il+5]+p1s[jl][il-5]*vls[jl][il-5])
			+coeffs[CX5]*(p1s[jl+5][il]*vls[jl+5][il]+p1s[jl-5][il]*vls[jl-5][il])
			+coeffs[CY5]*(p1y_p5*vy_p5+p1y_m5*vy_m5)
			+coeffs[CZ6]*(p1s[jl][il+6]*vls[jl][il+6]+p1s[jl][il-6]*vls[jl][il-6])
			+coeffs[CX6]*(p1s[jl+6][il]*vls[jl+6][il]+p1s[jl-6][il]*vls[jl-6][il])
			+coeffs[CY6]*(p1y_p6*vy_p6+p1y_m6*vy_m6)
			+coeffs[CZ7]*(p1s[jl][il+7]*vls[jl][il+7]+p1s[jl][il-7]*vls[jl][il-7])
			+coeffs[CX7]*(p1s[jl+7][il]*vls[jl+7][il]+p1s[jl-7][il]*vls[jl-7][il])
			+coeffs[CY7]*(p1y_p7*vy_p7+p1y_m7*vy_m7)
			+p1s[jl][il]+p1s[jl][il]-p0[addr];
		addr+=stride;
	}

}
