#define BLOCKZ_SIZE 16
#define BLOCKX_SIZE 16 
#define BLOCKY_SIZE 16
#define FAT 4
#define COEFFS_SIZE 13

#define C0  0
#define CZ1 1
#define CX1 2
#define CY1 3
#define CZ2 4
#define CX2 5
#define CY2 6
#define CZ3 7
#define CX3 8
#define CY3 9
#define CZ4 10
#define CX4 11
#define CY4 12

#define C_C00(d) (8.0/(5.0*(d)*(d)))
#define get_coeffs(d1,d2,d3) {-1025.0/576.0*(C_C00(d1)+C_C00(d2)+C_C00(d3)), C_C00(d1), C_C00(d2), C_C00(d3),-C_C00(d1)/8.0,-C_C00(d2)/8.0,-C_C00(d3)/8.0,C_C00(d1)/63.0,C_C00(d2)/63.0,C_C00(d3)/63.0,-C_C00(d1)/896.0,-C_C00(d2)/896.0,-C_C00(d3)/896.0}

__constant__ float coeffs[COEFFS_SIZE];

__global__ void wave_kernel_tex(int nblocksz, float *p0, float *p1, float *p2){
/* shared space */
	__shared__ float s_data[BLOCKX_SIZE+2*FAT][BLOCKZ_SIZE+2*FAT];

/* computing local(within thread block) and global(within global grid) coordinates */
	/* slow axis */
	int yblock=(blockIdx.x/nblocksz);
	int y0=FAT+yblock*BLOCKY_SIZE;
/* fast axis(z) and intermediate axis (x) */
	int zl=FAT+threadIdx.x; 				// local  zdir
	int zg=(blockIdx.x-yblock*nblocksz)*blockDim.x+zl; 	// global zdir
	int xl=FAT+threadIdx.y; 				// local  xdir
	int xg=blockIdx.y*blockDim.y+xl; 			// global xdir

/* variables keeping slow axis(y) points within a thread */
	float 	p1y_p1, p1y_m1,
		p1y_p2, p1y_m2,
		p1y_p3, p1y_m3,    
		p1y_p4, p1y_m4;

/* boundary condition for fast and intermediate axes */
	int bc_temp=min2(min2(bc_zmax-zg,bc_xmax-xg),xg); 

/* temp. variables for faster addressing */
	int stride=n1gpu*n2gpu;
	/* mem. address for current point */
	int addr=zg+n1gpu*xg+stride*y0;
	int addr_forward=addr-stride-stride-stride-stride;
//	int bc_agpu=30;
//	float bc_bgpu=0.0005;
	/* init. y-axis (slow) variables */
	p1y_m3=p1[addr_forward];
	p1y_m2=p1[addr_forward+=stride];
	p1y_m1=p1[addr_forward+=stride];
	s_data[xl][zl]=p1[addr_forward+=stride];
	p1y_p1=p1[addr_forward+=stride];
	p1y_p2=p1[addr_forward+=stride];
	p1y_p3=p1[addr_forward+=stride];
	p1y_p4=p1[addr_forward+=stride];

	/* looping over the slow y-axis */
	for(int yl=0; yl<BLOCKY_SIZE; ++yl){
		int y=y0+yl;

		p1y_m4=p1y_m3;		
		p1y_m3=p1y_m2;		
		p1y_m2=p1y_m1;
		p1y_m1=s_data[xl][zl];	
		s_data[xl][zl]=p1y_p1;
		p1y_p1=p1y_p2;		
		p1y_p2=p1y_p3;		
		p1y_p3=p1y_p4;
		p1y_p4=p1[addr_forward+=stride];

		/* boundary condition for slow axis */
		float bc=expf(bc_bgpu+min2(min2(y, bc_ymax-y),bc_temp)*bc_agpu);

/* loading into shared mem */
		if(threadIdx.x<FAT){ // load xdir fat/halo 
			s_data[xl][threadIdx.x]=p1[addr-FAT];
			s_data[xl][zl+BLOCKZ_SIZE]=p1[addr+BLOCKZ_SIZE];
		}
		if(threadIdx.y<FAT){ // load xdir fat/halo 
			s_data[threadIdx.y][zl]=p1[addr-FAT*n1gpu];
			s_data[xl+BLOCKX_SIZE][zl]=p1[addr+BLOCKX_SIZE*n1gpu];
		}
		__syncthreads();

//		int edge=min2(y-FAT,min2(min2(n2gpu-FAT-xg,n1gpu-FAT-zg),min2(n3gpu-FAT-y,xg-FAT)));
//		float temp=expf(-bc_bgpu*(bc_agpu-edge));

//		if(temp>1.) temp=1.;

		p2[addr]=
		tex3D(vel_ref,zg,xg,y)*
			(
			 coeffs[C0]*s_data[xl][zl]
			+coeffs[CZ1]*(s_data[xl][zl+1]+s_data[xl][zl-1])
			+coeffs[CX1]*(s_data[xl+1][zl]+s_data[xl-1][zl])
			+coeffs[CY1]*(p1y_p1+p1y_m1)
			+coeffs[CZ2]*(s_data[xl][zl+2]+s_data[xl][zl-2])
			+coeffs[CX2]*(s_data[xl+2][zl]+s_data[xl-2][zl])
			+coeffs[CY2]*(p1y_p2+p1y_m2)
			+coeffs[CZ3]*(s_data[xl][zl+3]+s_data[xl][zl-3])
			+coeffs[CX3]*(s_data[xl+3][zl]+s_data[xl-3][zl])
			+coeffs[CY3]*(p1y_p3+p1y_m3)
			+coeffs[CZ4]*(s_data[xl][zl+4]+s_data[xl][zl-4])
			+coeffs[CX4]*(s_data[xl+4][zl]+s_data[xl-4][zl])
			+coeffs[CY4]*(p1y_p4+p1y_m4)
			+bc*(p0[addr]-s_data[xl][zl])
		)+s_data[xl][zl]+s_data[xl][zl]-p0[addr];
		addr+=stride;
//		p1[addr]*=temp;
//		p0[addr]*=temp;
	}
}


extern "C" __global__ void damp_kernel(int nblocksz, float *p0, float *p1){
	int yblock=(blockIdx.x/nblocksz);
	int y0=FAT+yblock*BLOCKY_SIZE;
	int zl=FAT+threadIdx.x; 				// local  zdir
	int zg=(blockIdx.x-yblock*nblocksz)*blockDim.x+zl; 	// global zdir
	int xl=FAT+threadIdx.y; 				// local  xdir
	int xg=blockIdx.y*blockDim.y+xl; 			// global xdir	

	int stride=n1gpu*n2gpu;
	int bc_agpu=30;
	float bc_bgpu=0.0005;

	for(int yl=0; yl<BLOCKY_SIZE; ++yl){
		int y=y0+yl;
		int edge=min2( zg-FAT, min2(y-FAT,min2(min2(n2gpu-FAT-xg,n1gpu-FAT-zg),min2(n3gpu-FAT-y,xg-FAT))));
//		int edge=min2( zg-FAT, min2(y-FAT,min2(n1gpu-FAT-zg,min2(n3gpu-FAT-y,xg-FAT))));
		if(edge>=0){
		  int addr=zg+n1gpu*xg+stride*y;
		  float temp=expf(-bc_bgpu*(bc_agpu-edge));
		  if(temp<1.){
		    p1[addr]*=temp;
		    p0[addr]*=temp;
		  }
		}
	}
}
